/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ConnectException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.Types;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.http.HTTPAddress;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap12.SOAP12Address;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerException;

import org.jdom.JDOMException;
import org.jdom.input.DOMBuilder;
import org.jdom.input.SAXBuilder;
import org.jdom.output.DOMOutputter;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ibm.wsdl.xml.WSDLReaderImpl;
import com.ibm.wsdl.xml.WSDLWriterImpl;

/**
 * Util class to manipulate WSDL Document or Definition.
 * 
 * @author wjoseph - EBMWebsourcing
 * @since 1.0
 */
public final class WSDLHelper {

    public static String WSDL20_NS = "http://www.w3.org/ns/wsdl";

    public static String WSDL11_NS = "http://schemas.xmlsoap.org/wsdl/";

    public static String SCHEMA_NS = "http://www.w3.org/2001/XMLSchema";

    public static org.jdom.Namespace XSI = org.jdom.Namespace.getNamespace("xsi",
            "http://http://www.w3.org/2001/XMLSchema-instance");

    public static org.jdom.Namespace WSDL20 = org.jdom.Namespace.getNamespace(WSDL20_NS);

    public static org.jdom.Namespace WSDL11 = org.jdom.Namespace.getNamespace(WSDL11_NS);

    public static org.jdom.Namespace SCHEMA = org.jdom.Namespace.getNamespace(SCHEMA_NS);

    public static org.jdom.Namespace PETALS_WSDL_EXTENSIONS = org.jdom.Namespace.getNamespace(
            "petalsWsdlExtensions", "http://petals.ow2.org/wsdlExtensions");

    /**
     * Creates a new instance of {@link WSDLHelper}
     * 
     */
    private WSDLHelper() {
    }

    /**
     * Change the location attribute of the service node to
     * "/axis2/services/serviceName"
     * 
     * @param serviceName
     *            Name of the service
     * @param serviceDesc
     *            Dom representation of the service description
     */
    public static void changeServiceAddressInWSDL(final QName serviceName,
            final Definition serviceDef, final String prefix) {
        final SOAPAddress address = findServiceAddress(serviceDef, serviceName);
        if (address != null) {
            final String serviceAddress = prefix + serviceName.getLocalPart();
            address.setLocationURI(serviceAddress);
        }
    }

    /**
     * Change the location attribute of the service node to the given address
     * 
     * @param serviceName
     *            Name of the service
     * @param serviceDesc
     *            Dom representation of the service description
     * @param webServiceAddress
     *            new service address
     * @throws WSDLException
     * @throws TransformerException
     */
    public static Document changeServiceAddressInWSDL(final QName serviceName,
            final Document serviceDesc, final String webServiceAddress) throws WSDLException,
            TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        changeServiceAddressInWSDL(serviceName, serviceDef, webServiceAddress);
        return createDocumentFromDefinition(serviceDef);
    }

    /**
     * Creates a {@link Definition} from a {@link Document}
     * 
     * @param doc
     * @return
     * @throws WSDLException
     */
    public static Definition createDefinitionFromDocument(final Document doc) throws WSDLException {
        final WSDLReader reader = new WSDLReaderImpl();
        return reader.readWSDL(null, doc);
    }

    /**
     * Create a dom document from a File
     * 
     * @param wsdlFile
     *            file
     * @return dom document
     * @deprecated Use {@link WSDLHelper#createDocumentFromWSDL(File, boolean)}
     */
    @Deprecated
    public static Document createDocumentFromWSDL(final File wsdlFile) {
        URL url = null;
        Document doc = null;
        try {
            url = wsdlFile.toURI().toURL();
            doc = createDocumentFromWSDL(url);
        } catch (final Exception e) {
            e.printStackTrace();
        }

        return doc;
    }

    /**
     * Create a dom document from a File
     * 
     * @param wsdlFile
     *            file
     * @param wsdlImportsDownload
     *            If <code>true</code>, the external imports declared in the
     *            service WSDL will be downloaded, so they will be replaced by
     *            their content.
     * @return dom document
     */
    public static Document createDocumentFromWSDL(final File wsdlFile,
            final boolean wsdlImportsDownload) {
        URL url = null;
        Document doc = null;
        try {
            url = wsdlFile.toURI().toURL();
            doc = createDocumentFromWSDL(url, wsdlImportsDownload);
        } catch (final Exception e) {
            e.printStackTrace();
        }

        return doc;
    }

    /**
     * Get a {@link Document} from a wsdl located on the given {@link URL}.
     * 
     * @param url
     * @return null if there is an error getting the wsdl or if the parsing
     *         failed
     * @throws Exception
     *             the Document can not be created
     * @deprecated Use {@link WSDLHelper#createDocumentFromWSDL(URL, boolean)}
     */
    @Deprecated
    public static Document createDocumentFromWSDL(final URL url) throws PEtALSCDKException {
        return WSDLHelper.createDocumentFromWSDL(url, false);
    }

    /**
     * Get a {@link Document} from a wsdl located on the given {@link URL}.
     * 
     * @param url
     * @param wsdlImportsDownload
     *            If <code>false</code>, the external imports declared in the
     *            service WSDL won't be downloaded, so they won't be replaced by
     *            their content.
     * @return null if there is an error getting the wsdl or if the parsing
     *         failed
     * @throws Exception
     *             the Document can not be created
     */
    public static Document createDocumentFromWSDL(final URL url, final boolean wsdlImportsDownload)
            throws PEtALSCDKException {
        org.w3c.dom.Document doc = null;
        org.jdom.Document jdomDocument = null;
        URLConnection connection = null;
        InputStream is = null;

        try {
            connection = url.openConnection();
            is = connection.getInputStream();

            final SAXBuilder builder = new SAXBuilder();

            // Is it a local file ?
            String installroot = null;
            URI wsdlURI = url.toURI();
            if (wsdlURI.getScheme().equalsIgnoreCase("file")) {
                File wsdlFile = new File(wsdlURI);
                if (wsdlFile.exists()) {
                    File wsdlParent = new File(wsdlFile.getParent());
                    if (wsdlParent.isDirectory()) {
                        installroot = wsdlParent.getAbsolutePath();
                    }
                }
            }

            jdomDocument = builder.build(is);

            // add imported document
            jdomDocument = WSDLHelper.addImportedDocuments(jdomDocument, installroot,
                    wsdlImportsDownload);

            // Create DOM Document
            final DOMOutputter dom = new DOMOutputter();
            doc = dom.output(jdomDocument);

        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not get WSDL document for URL " + url, e);
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (final IOException e) {
                // Nothing to do
            }
        }
        return doc;
    }

    /**
     * Add the imported WSDl documents in the root document in the any part
     * 
     * @param doc
     * @param installroot
     * @param wsdlImportsDownload
     *            If <code>false</code>, the external imports declared in the
     *            service WSDL won't be downloaded, so they won't be replaced by
     *            their content.
     * @return
     * @throws PEtALSCDKException
     */
    @SuppressWarnings("unchecked")
    private static org.jdom.Document addImportedDocuments(final org.jdom.Document doc,
            final String installroot, final boolean wsdlImportsDownload) throws PEtALSCDKException {
        if (doc != null) {
            final org.jdom.Element root = doc.getRootElement();
            final String wsdlRoot = root.getName();
            if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                    || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                            .equals(WSDL20_NS)))) {

                final List<org.jdom.Element> imports = WSDLHelper.getAllImportedDocuments(doc,
                        installroot, wsdlImportsDownload);
                if ((imports != null) && (imports.size() > 0)) {
                    root.addNamespaceDeclaration(PETALS_WSDL_EXTENSIONS);
                    root.addNamespaceDeclaration(XSI);

                    if (root.getAttribute("schemaLocation", XSI) == null) {
                        root.setAttribute("schemaLocation",
                                "http://petals.ow2.org/wsdlExtensions wsdlExtensions.xsd", XSI);
                    } else {
                        String val = root.getAttributeValue("schemaLocation", XSI);
                        val = val + " " + "http://petals.ow2.org/wsdlExtensions wsdlExtensions.xsd";
                        root.setAttribute("schemaLocation", val, XSI);
                    }

                    final org.jdom.Element importedDocumentsElmt = new org.jdom.Element(
                            "importedDocuments", PETALS_WSDL_EXTENSIONS);
                    for (final org.jdom.Element impt : imports) {
                        importedDocumentsElmt.addContent(impt);
                    }
                    root.getChildren().add(importedDocumentsElmt);
                }
            }
        }
        return doc;
    }

    /**
     * 
     * @param doc The master wsdl document to import
     * @param installroot The root master wsdl document path 
     * @param wsdlImportsDownload If true, the imported document (wsdl or xsd)<br>
     * mut be imported too. If false, they are ignored
     * @return a list of documents imported
     * @throws PEtALSCDKException
     */
    private static List<org.jdom.Element> getAllImportedDocuments(final org.jdom.Document doc,
            final String installroot, final boolean wsdlImportsDownload) throws PEtALSCDKException {
        List<org.jdom.Element> res = null;
        String importedDocumentInstallroot = null;
        if (doc != null) {
            final org.jdom.Element root = doc.getRootElement();
            final String wsdlRoot = root.getName();
            if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                    || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                            .equals(WSDL20_NS)))
                    || ((wsdlRoot.equals("schema")) && (root.getNamespaceURI().equals(SCHEMA_NS)))) {

                final List<org.jdom.Element> imports = getAllImports(root);

                // get in shemaLocation
                if (root.getAttributeValue("schemaLocation", XSI) != null) {
                    final String val = root.getAttributeValue("schemaLocation", XSI);
                    final StringTokenizer st = new StringTokenizer(val, " ");
                    String ns = null;
                    String location = null;
                    org.jdom.Element e = null;
                    while (st.hasMoreTokens()) {
                        ns = st.nextToken();
                        if (st.hasMoreTokens()) {
                            location = st.nextToken();
                            e = new org.jdom.Element("import", WSDL20);
                            e.setAttribute("schemaLocation", location);
                            e.setAttribute("namespace", ns);
                            imports.add(e);
                        }
                    }
                }

                if (imports.size() > 0) {
                    res = new ArrayList<org.jdom.Element>();
                    List<org.jdom.Element> inside = null;
                    String name = null;
                    String uri = null;
                    SAXBuilder builder = null;
                    org.jdom.Document importedDocument = null;
                    org.jdom.Element documentElmt = null;
                    URLConnection connection = null;
                    InputStream is = null;
                    try {
                        for (final org.jdom.Element impt : imports) {
                            name = impt.getAttributeValue("location");
                            if (name == null) {
                                name = impt.getAttributeValue("schemaLocation");
                            }
                            if (name != null) {
                                boolean isExternalURL = false;
                                try {
                                    final URL nameURL = new URL(name);
                                    final String protocol = nameURL.getProtocol();
                                    if ("http".equalsIgnoreCase(protocol)
                                            || "https".equalsIgnoreCase(protocol)) {
                                        isExternalURL = true;
                                    } else {
                                        isExternalURL = false;
                                    }
                                } catch (final MalformedURLException e1) {
                                    // The schemaLocation value is not an
                                    // external
                                    // URL because it is not wellformed
                                    isExternalURL = false;
                                }

                                if (!isExternalURL || (isExternalURL && wsdlImportsDownload)) {

                                    try {
                                        if (installroot != null) {
                                            uri = installroot + File.separator + name;
                                            final File f = new File(uri);
                                            importedDocumentInstallroot = f.getCanonicalFile()
                                                    .getParent();
                                            if (f.exists()) {
                                                connection = f.toURI().toURL().openConnection();
                                            } else {
                                                connection = new URI(name).toURL().openConnection();
                                            }
                                        } else {
                                            importedDocumentInstallroot = null;
                                            connection = new URI(name).toURL().openConnection();
                                        }
                                        is = connection.getInputStream();

                                        builder = new SAXBuilder();
                                        importedDocument = builder.build(is);

                                        if (importedDocument != null) {
                                            documentElmt = new org.jdom.Element("document",
                                                    PETALS_WSDL_EXTENSIONS);
                                            documentElmt.setAttribute("name", name);

                                            documentElmt
                                                    .addContent((org.jdom.Element) importedDocument
                                                            .getRootElement().clone());

                                            res.add(documentElmt);

                                            inside = WSDLHelper.getAllImportedDocuments(
                                                    importedDocument, importedDocumentInstallroot,
                                                    wsdlImportsDownload);

                                            if (inside != null) {
                                                for (final org.jdom.Element document : inside) {
                                                    res.add(document);
                                                }
                                            }
                                        }
                                    } catch (final MalformedURLException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final URISyntaxException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final UnknownHostException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final ConnectException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                        // TODO: This exception occurs when the
                                        // import is
                                        // an XSD available through an HTTP
                                        // connection and its
                                        // server is unacessible.
                                        // To prevent the server unacessibility
                                        // beacuse Internet
                                        // proxy, following parameters should be
                                        // added to the
                                        // CDK configuration: proxy host, proxy
                                        // port and non
                                        // proxied hosts (see feature#311110 on
                                        // the forge)
                                    }
                                }
                            }
                        }
                    } catch (final IOException e) {
                        throw new PEtALSCDKException(
                                "Can not get imported document for URL " + uri, e);
                    } catch (final JDOMException e) {
                        throw new PEtALSCDKException(
                                "Can not get imported document for URL " + uri, e);
                    }
                }
            }
        }
        return res;
    }

    /**
     * 
     * @param doc
     * @return
     * @throws PEtALSCDKException
     */
    @SuppressWarnings("unchecked")
    public static List<Element> getImportedDocuments(Document doc) throws PEtALSCDKException {
        final List<Element> importedDocs = new ArrayList<Element>();
        try {
            if (doc != null) {
                // Create JDOM Document
                // Create new DOMBuilder, using default parser
                final DOMBuilder builder = new DOMBuilder();
                final org.jdom.Document jdomDoc = builder.build(doc);

                final org.jdom.Element root = jdomDoc.getRootElement();

                final org.jdom.Element importedDocumentsElmt = root.getChild("importedDocuments",
                        PETALS_WSDL_EXTENSIONS);

                // Create DOM Document
                final DOMOutputter dom = new DOMOutputter();

                if (importedDocumentsElmt != null) {

                    boolean find = false;
                    org.jdom.Element description = null;
                    for (final org.jdom.Element element : (List<org.jdom.Element>) importedDocumentsElmt
                            .getChildren()) {

                        description = element.getChild("description", WSDL20);
                        if (description == null) {
                            description = element.getChild("definitions", WSDL11);
                        }

                        if (description != null) {
                            for (final org.jdom.Namespace nsRoot : (List<org.jdom.Namespace>) root
                                    .getAdditionalNamespaces()) {
                                find = false;
                                for (final org.jdom.Namespace nsElmt : (List<org.jdom.Namespace>) description
                                        .getAdditionalNamespaces()) {
                                    if ((nsRoot.getPrefix().equals(nsElmt.getPrefix()))
                                            && (nsRoot.equals(nsElmt))) {
                                        find = true;
                                    }
                                }
                                if (!find) {
                                    find = false;
                                    for (final org.jdom.Namespace nsElmt : (List<org.jdom.Namespace>) description
                                            .getAdditionalNamespaces()) {
                                        if (nsElmt.getPrefix().equals(nsRoot.getPrefix())) {
                                            find = true;
                                        }
                                    }

                                    if (!find) {
                                        description.addNamespaceDeclaration(nsRoot);
                                    }
                                }
                            }
                        }
                        importedDocs.add(dom.output(element));
                    }

                }
                root.removeContent(importedDocumentsElmt);
                doc = dom.output(jdomDoc);
            }
        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not extract imported document in WSDL", e);
        }
        return importedDocs;
    }

    /**
     * 
     * @param doc
     * @return
     * @throws PEtALSCDKException
     */
    public static Document removeImportedDocuments(Document doc) throws PEtALSCDKException {
        try {
            if (doc != null) {
                // Create JDOM Document
                // Create new DOMBuilder, using default parser
                final DOMBuilder builder = new DOMBuilder();
                final org.jdom.Document jdomDoc = builder.build(doc);

                final org.jdom.Element root = jdomDoc.getRootElement();
                final String wsdlRoot = root.getName();
                if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                        || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                                .equals(WSDL20_NS)))) {

                    if (root.getAdditionalNamespaces().contains(PETALS_WSDL_EXTENSIONS)) {
                        root.removeNamespaceDeclaration(PETALS_WSDL_EXTENSIONS);

                        String val = root.getAttributeValue("schemaLocation", XSI);
                        if (val != null) {
                            val = val.replace(
                                    "http://petals.ow2.org/wsdlExtensions wsdlExtensions.xsd", "");
                            if (val.trim().length() == 0) {
                                root.removeAttribute("schemaLocation", XSI);
                            }
                        }
                        final org.jdom.Element importedDocumentsElmt = root.getChild(
                                "importedDocuments", PETALS_WSDL_EXTENSIONS);

                        if (importedDocumentsElmt != null) {
                            root.removeContent(importedDocumentsElmt);
                            // Create DOM Document
                            final DOMOutputter dom = new DOMOutputter();
                            doc = dom.output(jdomDoc);
                        }
                    }
                }
            }
        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not remove imported documents in WSDL", e);
        }
        return doc;
    }

    /**
     * 
     * @param doc
     * @param path
     * @return
     * @throws PEtALSCDKException
     */
    public static Document changeImportLocation(Document doc, final String path)
            throws PEtALSCDKException {
        try {
            if (doc != null) {
                // Create JDOM Document
                // Create new DOMBuilder, using default parser
                final DOMBuilder builder = new DOMBuilder();
                final org.jdom.Document jdomDoc = builder.build(doc);

                final org.jdom.Element root = jdomDoc.getRootElement();
                final String wsdlRoot = root.getName();
                if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                        || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                                .equals(WSDL20_NS)))
                        || ((wsdlRoot.equals("schema")) && (root.getNamespaceURI()
                                .equals(SCHEMA_NS)))) {

                    final List<org.jdom.Element> imports = getAllImports(root);

                    // get in shemaLocation
                    if (root.getAttributeValue("schemaLocation", XSI) != null) {
                        String val = root.getAttributeValue("schemaLocation", XSI);
                        final StringTokenizer st = new StringTokenizer(val, " ");
                        String location = null;
                        String value = null;
                        while (st.hasMoreTokens()) {
                            st.nextToken();
                            value = null;
                            if (st.hasMoreTokens()) {
                                location = st.nextToken();
                                boolean isUrl = true;
                                try {
                                    new URL(location);
                                } catch (MalformedURLException e) {
                                    isUrl = false;
                                }
                                if (!isUrl) {
                                    value = path + location;
                                    val = val.replace(location, value);
                                }
                            }
                        }
                        root.getAttribute("schemaLocation", XSI).setValue(val);
                    }

                    if ((imports != null) && (imports.size() > 0)) {
                        for (final org.jdom.Element impt : imports) {
                            String value = impt.getAttributeValue("location");
                            if (value == null) {
                                value = impt.getAttributeValue("schemaLocation");
                            }

                            if (value != null) {
                                value = path + value;
                                if (impt.getAttribute("location") != null) {
                                    boolean isUrl = true;
                                    try {
                                        new URL(impt.getAttribute("location").getValue());
                                    } catch (MalformedURLException e) {
                                        isUrl = false;
                                    }
                                    if (!isUrl) {
                                        impt.getAttribute("location").setValue(value);
                                    }
                                } else if (impt.getAttribute("schemaLocation") != null) {
                                    boolean isUrl = true;
                                    try {
                                        new URL(impt.getAttribute("schemaLocation").getValue());
                                    } catch (MalformedURLException e) {
                                        isUrl = false;
                                    }
                                    if (!isUrl) {
                                        impt.getAttribute("schemaLocation").setValue(value);
                                    }
                                }
                            }
                        }

                        // Create DOM Document
                        final DOMOutputter dom = new DOMOutputter();
                        doc = dom.output(jdomDoc);
                    }
                }
            }
        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not change import location in documents in WSDL", e);
        }
        return doc;
    }

    /**
     * Get all the imports from the root element
     * 
     * @param root
     * @return
     */
    @SuppressWarnings("unchecked")
    private static List<org.jdom.Element> getAllImports(final org.jdom.Element root) {
        final List<org.jdom.Element> imports = new ArrayList<org.jdom.Element>();

        // get in wsdl import or include
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("import",
                WSDL11)) {
            imports.add((org.jdom.Element) impt);
        }
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("import",
                WSDL20)) {
            imports.add((org.jdom.Element) impt);
        }
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("include",
                WSDL20)) {
            imports.add((org.jdom.Element) impt);
        }

        // get in schema
        org.jdom.Element types = root.getChild("types", WSDL11);
        org.jdom.Element schema = null;
        if (types != null) {
            schema = types.getChild("schema", SCHEMA);
        } else {
            types = root.getChild("types", WSDL20);
            if (types != null) {
                schema = types.getChild("schema", SCHEMA);
            }
        }

        if (schema != null) {
            for (final org.jdom.Element impt : (List<org.jdom.Element>) schema.getChildren(
                    "import", SCHEMA)) {
                imports.add((org.jdom.Element) impt);
            }
            for (final org.jdom.Element impt : (List<org.jdom.Element>) schema.getChildren(
                    "include", SCHEMA)) {
                imports.add((org.jdom.Element) impt);
            }
        }
        return imports;
    }

    /**
     * Find the service address for the given service name.
     * 
     * @param serviceDef
     *            the WSDL definition
     * @param serviceName
     *            the qualified service name
     * @return {@link SOAPAddress} if service found in definition, else null
     */
    @SuppressWarnings("unchecked")
    public static SOAPAddress findServiceAddress(final Definition serviceDef,
            final QName serviceName) {
        SOAPAddress result = null;
        if (serviceDef != null) {
            final Service service = serviceDef.getService(serviceName);
            if (service != null) {
                final Map<String, Port> ports = service.getPorts();
                for (final Port port : ports.values()) {
                    final List<ExtensibilityElement> extensibilityElements = port
                            .getExtensibilityElements();
                    for (final ExtensibilityElement element : extensibilityElements) {
                        // The SOAP binding is an extensibility element of
                        // Port
                        if (element instanceof SOAPAddress) {
                            result = (SOAPAddress) element;
                            break;
                        }
                    }
                    if (result != null) {
                        break;
                    }
                }
            }
        }
        return result;
    }

    /**
     * Returns the service's port name contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service port name attribute or null if not found
     */
    @SuppressWarnings("unchecked")
    public static String getPortNameFromWSDLDefinition(final Definition serviceDef,
            final QName serviceName) {
        String result = null;
        final Service service = serviceDef.getService(serviceName);
        final Map<String, Port> ports = service.getPorts();
        for (final Port port : ports.values()) {
            result = port.getName();
        }
        return result;
    }

    /**
     * Returns the service's port name contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service port name attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static String getPortNameFromWSDLDocument(final Document serviceDesc,
            final QName serviceName) throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getPortNameFromWSDLDefinition(serviceDef, serviceName);
    }

    /**
     * Returns the service address contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service address attribute or null if not found
     */
    public static String getServiceAddressFromWSDLDefinition(final Definition serviceDef,
            final QName serviceName) {
        String result = null;
        final SOAPAddress address = findServiceAddress(serviceDef, serviceName);
        if (address != null) {
            result = address.getLocationURI();
        }
        return result;
    }

    /**
     * Returns the service address contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service address attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static String getServiceAddressFromWSDLDocument(final Document serviceDesc,
            final QName serviceName) throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getServiceAddressFromWSDLDefinition(serviceDef, serviceName);
    }

    /**
     * Returns the service address contained in a WSDL 1.0 file
     * 
     * @param wsdlFile
     *            The WSDL file
     * @return The service address attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static String getServiceAddressFromWSDLFile(final File wsdlFile, final QName serviceName)
            throws WSDLException, TransformerException {
        final Document doc = createDocumentFromWSDL(wsdlFile);
        return getServiceAddressFromWSDLDocument(doc, serviceName);
    }

    /**
     * Returns the service name contained in a WSDL 1.1 dom Document
     * 
     * @param serviceDef
     *            The dom Document
     * @return The service name attribute or null if not found
     */
    @SuppressWarnings("unchecked")
    public static QName[] getServiceNameFromWSDLDefinition(final Definition serviceDef) {
        QName[] result = null;
        if (serviceDef != null) {
            final Map<String, Service> services = serviceDef.getServices();
            result = new QName[services.size()];
            int i = 0;
            for (final Service service : services.values()) {
                result[i++] = service.getQName();
            }
        }
        return result;
    }

    /**
     * Returns the service name contained in a WSDL 1.1 dom Document
     * 
     * @param serviceDef
     *            The dom Document
     * @return The service name attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static QName[] getServiceNameFromWSDLDocument(final Document serviceDesc)
            throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getServiceNameFromWSDLDefinition(serviceDef);
    }

    /**
     * Returns the service name contained in a WSDL 1.0 file
     * 
     * @param wsdlFile
     *            The WSDL file
     * @return The service name attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static QName[] getServiceNameFromWSDLFile(final File wsdlFile) throws WSDLException,
            TransformerException {
        final Document doc = createDocumentFromWSDL(wsdlFile);
        return getServiceNameFromWSDLDocument(doc);
    }

    /**
     * Returns WSDL 1.0 document target namespace
     * 
     * @param serviceDesc
     *            The service description document
     * @return The target namespace or null if not found
     */
    public static String getTargetNSFromWSDLDefinition(final Definition serviceDef) {
        return serviceDef.getTargetNamespace();
    }

    /**
     * Returns WSDL 1.0 document target namespace
     * 
     * @param serviceDesc
     *            The service description document
     * @return The target namespace or null if not found
     * @throws WSDLException
     * @throws TransformerException
     */
    public static String getTargetNSFromWSDLDocument(final Document serviceDesc)
            throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getTargetNSFromWSDLDefinition(serviceDef);
    }

    /**
     * Search the service description for an operation
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found or else false
     */
    public static boolean hasOperationNamed(final Definition serviceDef, final String opName,
            final QName serviceName) {
        final BindingOperation op = findOperationNamed(serviceDef, opName, serviceName);
        return op != null;
    }

    /**
     * Search the service description for an operation
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found or else false
     * @throws WSDLException
     * @throws TransformerException
     */
    public static boolean hasOperationNamed(final Document serviceDesc, final String opName,
            final QName serviceName) throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return hasOperationNamed(serviceDef, opName, serviceName);
    }

    /**
     * Search the service description for an operation. If the operation is
     * found , checks if it has a child node named 'output'
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found and has the output child
     *         node or else false
     */
    public static boolean isInOutOperation(final Definition serviceDef, final String opName,
            final QName serviceName) {
        boolean inOut = false;
        final BindingOperation op = findOperationNamed(serviceDef, opName, serviceName);
        if (op != null) {
            inOut = op.getBindingOutput() != null;
        }
        return inOut;
    }

    /**
     * Search the service description for an operation. If the operation is
     * found , checks if it has a child node named 'output'
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found and has the output child
     *         node or else false
     * @throws WSDLException
     * @throws TransformerException
     */
    public static boolean isInOutOperation(final Document serviceDesc, final String opName,
            final QName serviceName) throws WSDLException, TransformerException {
        final String docString = XMLUtil.createStringFromDOMDocument(serviceDesc);
        final Document tmpDoc = XMLUtil.createDocumentFromString(docString);
        final Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return isInOutOperation(serviceDef, opName, serviceName);
    }

    /**
     * Create a {@link Document} from a {@link Definition}
     * 
     * @param def
     * @return
     * @throws WSDLException
     */
    protected static Document createDocumentFromDefinition(final Definition def)
            throws WSDLException {
        Document result = null;
        final WSDLWriter writer = new WSDLWriterImpl();
        final OutputStream outStream = new ByteArrayOutputStream();
        writer.writeWSDL(def, outStream);
        final String documentToString = outStream.toString();
        result = XMLUtil.createDocumentFromString(documentToString);
        return result;
    }

    /**
     * 
     * @param serviceDef
     * @param opName
     * @param serviceName
     * @return
     */
    @SuppressWarnings("unchecked")
    protected static BindingOperation findOperationNamed(final Definition serviceDef,
            final String opName, final QName serviceName) {
        BindingOperation result = null;
        boolean found = false;
        if (serviceDef != null) {

            final Service service = findService(serviceDef, serviceName);
            if (service != null) {
                final Map<String, Port> ports = service.getPorts();
                for (final Port port : ports.values()) {
                    final List<BindingOperation> ops = port.getBinding().getBindingOperations();
                    for (final BindingOperation op : ops) {
                        if (op.getName().equals(opName)) {
                            result = op;
                            found = true;
                            break;
                        }
                    }
                    if (found) {
                        break;
                    }
                }
            }
        }
        return result;
    }

    /**
     * Return the Service definition for this service. This method is comparable
     * to definition.getService(qname_service), but due to some variation in
     * QName structure, we perform manually a equals(qname_service, qname) to
     * retreive the service description
     * 
     * @param definition
     *            wsdl definition, can be null
     * @param service
     *            service to retrieve its definition, can be null
     * @return null if definition or service are null, or service not found in
     *         description
     */
    public static Service findService(final Definition definition, final QName service) {
        Service result = null;

        if ((definition != null) && (service != null)) {
            for (final Iterator<?> iter = definition.getServices().keySet().iterator(); iter
                    .hasNext()
                    && (result == null);) {
                final QName element = (QName) iter.next();
                if (equals(element, service)) {
                    result = definition.getService(element);
                }

            }
        }
        return result;
    }

    /**
     * Perform a qname.toString equality
     * 
     * @param a
     *            can be null
     * @param b
     *            can be null
     * @return a.toString equals b.toString, or a=b=null
     */
    public static boolean equals(final QName a, final QName b) {
        boolean result = true;

        if ((a != null) && (b != null)) {
            result = a.toString().equals(b.toString());
        } else {
            result = (a == null) && (b == null);
        }
        return result;
    }

    /**
     * Create a light WSDL2.0 document from the given parameters. All parameters
     * are required, otherwise a NullPointerException will be thrown. Namespaces
     * are managed.
     * 
     * @param interfaceName
     *            must be non null
     * @param serviceName
     *            must be non null
     * @param endpointName
     *            must be non null
     * @return a WSDL2.0 light document
     */
    public static Document createLightWSDL20(final QName interfaceName, final QName serviceName,
            final String endpointName) {

        final StringBuffer s = new StringBuffer();

        // retrieve the namespace if defined
        String targetNamespace = interfaceName.getNamespaceURI();
        if (StringHelper.isNullOrEmpty(interfaceName.getNamespaceURI())) {
            targetNamespace = serviceName.getNamespaceURI();
        }

        // write definition xmlns block
        s.append("<?xml version='1.0' encoding='UTF-8'?>\n");
        s.append("  <description ");
        if (!StringHelper.isNullOrEmpty(targetNamespace)) {
            s.append(" targetNamespace='" + targetNamespace + "'");
            s.append(" xmlns:tns='" + targetNamespace + "'");
        }
        s.append(" xmlns='http://www.w3.org/2006/01/wsdl'>\n");

        // write interface block
        s.append("    <interface name='" + interfaceName.getLocalPart() + "'></interface>\n");

        // write binding block
        s.append("    <binding name='" + endpointName + "'></binding>\n");

        // write service block (with interface and endpoint part)
        s.append("    <service name='" + serviceName.getLocalPart() + "'\n");
        if (!StringHelper.isNullOrEmpty(targetNamespace)) {
            s.append("             interface='tns:" + interfaceName.getLocalPart() + "'>\n");
            s.append("      <endpoint binding='tns:" + endpointName + "'></endpoint>\n");
        } else {
            s.append("             interface='" + interfaceName.getLocalPart() + "'>\n");
            s.append("      <endpoint binding='" + endpointName + "'></endpoint>\n");
        }

        s.append("    </service>\n");
        s.append("  </description>");

        // construct the document
        final ByteArrayInputStream is = new ByteArrayInputStream(s.toString().getBytes());

        Document doc = null;
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        try {
            final DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(is);
        } catch (final Exception e) {
            // this kind of exception should never be thrown...
        }
        return doc;
    }

    /**
     * Change all the addresses prefix of the WSDL document to the new one.
     * 
     * @param doc
     * @param newPrefix
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Document changeAdressPrefixInWSDL(final Document doc, final String newPrefix)
            throws Exception {

        final WSDLFactory wsdlFactory = WSDLFactory.newInstance();
        final WSDLReader reader = wsdlFactory.newWSDLReader();
        final Definition def = reader.readWSDL(null, doc);

        final Map services = def.getAllServices();
        final Iterator iter = services.keySet().iterator();
        while (iter.hasNext()) {
            final Service service = (Service) services.get(iter.next());

            final Map<String, Port> ports = service.getPorts();
            for (final Port port : ports.values()) {
                final List<ExtensibilityElement> extensibilityElements = port
                        .getExtensibilityElements();

                for (final ExtensibilityElement element : extensibilityElements) {
                    if (element instanceof SOAPAddress) {
                        final SOAPAddress address = (SOAPAddress) element;
                        address.setLocationURI(newPrefix + "/" + service.getQName().getLocalPart());

                    } else if (element instanceof SOAP12Address) {
                        final SOAP12Address address = (SOAP12Address) element;
                        address.setLocationURI(newPrefix + "/" + service.getQName().getLocalPart());

                    } else if (element instanceof HTTPAddress) {
                        final HTTPAddress address = (HTTPAddress) element;
                        address.setLocationURI(newPrefix + "/" + service.getQName().getLocalPart());
                    }
                }
            }
        }
        return createDocumentFromDefinition(def);
    }

    /**
     * Replace the wsdl ports address with the given one
     * 
     * @param doc
     * @param localAddress
     * @return
     * @throws TransformerException
     */
    public static Document replaceAddressPrefix(final Document doc, final String localAddress)
            throws TransformerException {

        // get services
        final NodeList services = doc.getElementsByTagName("wsdl:service");
        for (int i = 0; i < services.getLength(); i++) {
            final Element serviceElement = (Element) services.item(i);
            final String serviceName = serviceElement.getAttribute("name");
            final NodeList ports = serviceElement.getElementsByTagName("wsdl:port");

            // get the ports
            for (int j = 0; j < ports.getLength(); j++) {
                final Element portElement = (Element) ports.item(j);

                // get addresses
                NodeList addresses = portElement.getElementsByTagName("soap:address");

                for (int k = 0; k < addresses.getLength(); k++) {
                    final Element addr = (Element) addresses.item(k);
                    addr.setAttribute("location", localAddress + "/" + serviceName);
                }
                addresses = portElement.getElementsByTagName("soap12:address");
                for (int k = 0; k < addresses.getLength(); k++) {
                    final Element addr = (Element) addresses.item(k);
                    addr.setAttribute("location", localAddress + "/" + serviceName);
                }
                addresses = portElement.getElementsByTagName("http:address");
                for (int k = 0; k < addresses.getLength(); k++) {
                    final Element addr = (Element) addresses.item(k);
                    addr.setAttribute("location", localAddress + "/" + serviceName);
                }
            }
        }
        final String docString = XMLUtil.createStringFromDOMDocument(doc);
        return XMLUtil.createDocumentFromString(docString);
    }

    /**
     * Replace all the addresses by the given one. Ignore all the name spaces.
     * 
     * @param doc
     * @param localAddress
     * @return
     * @throws TransformerException
     */
    public static Document replaceAddressPrefixIgnoreNS(final Document doc,
            final String localAddress) throws TransformerException {
        final List<Node> services = WSDLHelper.getChildNodes(doc.getFirstChild(), "service");
        for (final Node service : services) {
            String serviceName = null;
            if (service instanceof Element) {
                final Element e = (Element) service;
                serviceName = e.getAttribute("name");
            }

            final List<Node> ports = WSDLHelper.getChildNodes(service, "port");

            for (final Node port : ports) {
                final List<Node> addresses = WSDLHelper.getChildNodes(port, "address");
                for (final Node address : addresses) {
                    if (address instanceof Element) {
                        final Element a = (Element) address;
                        a.setAttribute("location", localAddress + "/" + serviceName);
                    }
                }
            }
        }
        final String docString = XMLUtil.createStringFromDOMDocument(doc);
        return XMLUtil.createDocumentFromString(docString);
    }

    /**
     * Replace the service address in the wsdl:address element, not in the
     * ports.
     * 
     * @param doc
     * @param localAddress
     * @param newServiceName
     * @return
     * @throws TransformerException
     */
    public static Document replaceAddressServiceIgnoreNS(final Document doc,
            final String localAddress, final String newServiceName) throws TransformerException {
        final List<Node> services = WSDLHelper.getChildNodes(doc.getFirstChild(), "service");
        for (final Node service : services) {
            final List<Node> ports = WSDLHelper.getChildNodes(service, "port");

            for (final Node port : ports) {
                final List<Node> addresses = WSDLHelper.getChildNodes(port, "address");
                for (final Node address : addresses) {
                    if (address instanceof Element) {
                        final Element a = (Element) address;
                        a.setAttribute("location", localAddress + "/" + newServiceName);
                    }
                }
            }
        }
        final String docString = XMLUtil.createStringFromDOMDocument(doc);
        return XMLUtil.createDocumentFromString(docString);
    }

    /**
     * Replace the address
     * 
     * @param doc
     * @param localAddress
     * @return
     * @throws TransformerException
     */
    public static Document replaceAddressAndServiceIgnoreNS(final Document doc,
            final String newAddress, final String newServiceName) throws TransformerException {
        final List<Node> services = WSDLHelper.getChildNodes(doc.getFirstChild(), "service");
        for (final Node service : services) {
            if (service instanceof Element) {
                final Element e = (Element) service;
                if (e.getAttribute("name") != null) {
                    try {
                        e.setAttribute("name", newServiceName);
                    } catch (final DOMException e1) {
                    }
                }
            }

            final List<Node> ports = WSDLHelper.getChildNodes(service, "port");

            for (final Node port : ports) {
                final List<Node> addresses = WSDLHelper.getChildNodes(port, "address");
                for (final Node address : addresses) {
                    if (address instanceof Element) {
                        final Element a = (Element) address;
                        a.setAttribute("location", newAddress + "/" + newServiceName);
                    }
                }
            }
        }
        final String docString = XMLUtil.createStringFromDOMDocument(doc);
        return XMLUtil.createDocumentFromString(docString);
    }

    /**
     * Get the children of the current with name ending with the given one
     * (ignore NS)
     * 
     * @param element
     * @param name
     * @return
     */
    public static List<Node> getChildNodes(final Node element, final String name) {
        final List<Node> elements = new ArrayList<Node>();
        final NodeList children = element.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            final Node tmp = children.item(i);
            if (tmp.getNodeName().endsWith(name)) {
                elements.add(tmp);
            }
        }
        return elements;
    }

    /**
     * Print the WSDL to the output stream
     * 
     * @param doc
     * @param out
     */
    public static void printWSDL(final Document doc, final OutputStream out) {
        try {
            final String buffer = XMLUtil.createStringFromDOMDocument(doc);
            out.write(buffer.getBytes());
        } catch (final IOException e) {
            e.printStackTrace();
        } catch (final TransformerException e) {
            e.printStackTrace();
        }
    }

    /**
     * Process the WSDL imports of the WSDL4J definition. The definition will be
     * updated with all the imports. If it is serialized after this processing,
     * the WSDL imports will be included in the definition body to create a
     * single WSDL document without imports.
     * 
     * @param definition
     * @param processedDocuments
     */
    @SuppressWarnings("unchecked")
    public static void processWSDLImports(final Definition definition,
            final List<String> processedDocuments) {

        final Map wsdlImports = definition.getImports();

        if ((null != wsdlImports) && !wsdlImports.isEmpty()) {
            final Collection<?> importsCollection = wsdlImports.values();
            for (final Iterator iterator = importsCollection.iterator(); iterator.hasNext();) {
                final Vector values = (Vector) iterator.next();
                for (int i = 0; i < values.size(); i++) {
                    final Import wsdlImport = (Import) values.elementAt(i);

                    if (wsdlImport.getDefinition() != null) {
                        final Definition importedDef = wsdlImport.getDefinition();

                        if (importedDef != null) {
                            String key = importedDef.getDocumentBaseURI();
                            if (key == null) {
                                key = importedDef.getTargetNamespace();
                            }
                            // stop recursive imports!
                            if (processedDocuments.contains(key)) {
                                return;
                            }
                            processedDocuments.add(key);

                            processWSDLImports(importedDef, processedDocuments);

                            // copy ns
                            final Map namespaces = importedDef.getNamespaces();
                            final Iterator<String> keys = namespaces.keySet().iterator();
                            while (keys.hasNext()) {
                                final String key2 = keys.next();
                                if (!definition.getNamespaces().containsValue(namespaces.get(key2))) {

                                    // prefix is not defined in parent
                                    // definition, create new one
                                    int id = 0;
                                    String prefix = key2;
                                    while (definition.getNamespace(prefix) != null) {
                                        prefix = key2 + id;
                                        id++;
                                    }
                                    definition
                                            .addNamespace(prefix, namespaces.get(key2).toString());
                                }
                            }

                            // copy types
                            final Types t = importedDef.getTypes();
                            if (t != null) {
                                final List typesList = t.getExtensibilityElements();
                                for (int j = 0; j < typesList.size(); j++) {
                                    Types types = definition.getTypes();
                                    if (types == null) {
                                        types = definition.createTypes();
                                        definition.setTypes(types);
                                    }
                                    types.addExtensibilityElement((ExtensibilityElement) typesList
                                            .get(j));
                                }
                            }

                            // add messages
                            final Map messagesMap = importedDef.getMessages();
                            definition.getMessages().putAll(messagesMap);

                            // add port types
                            final Map porttypeMap = importedDef.getPortTypes();
                            definition.getPortTypes().putAll(porttypeMap);

                            // add bindings
                            final Map bindingMap = importedDef.getBindings();
                            definition.getBindings().putAll(bindingMap);

                            // add services
                            final Map serviceMap = importedDef.getServices();
                            definition.getServices().putAll(serviceMap);

                            final List extElementList = importedDef.getExtensibilityElements();
                            definition.getExtensibilityElements().addAll(extElementList);

                        }
                    }
                    // remove the import section from the parent definition
                    definition.removeImport(wsdlImport);
                }
            }
        }
    }
}
